(* ::Package:: *)

(* ::Section:: *)
(*Macro implementation*)


BeginPackage["MacroProcessor`"];


Define\[Breve]Macro::usage="Define\[Breve]Macro[patt,val] \[Dash] Defines a macro that replaces patt with val before execution starts. Define\[Breve]Macro[symbol] \[Dash] Defines an executable macro with the given head";
Macro\[Breve]Hold::usage="Executable macros often wrap their final answer in Macro\[Breve]Hold to prevent evaluation until after re-insertion into the code.";


Performing\[Breve]Macro\[Breve]Expansion::usage="Performing\[Breve]Macro\[Breve]Expansion[] \[Dash] Callable from within an executable macro to determine if it is being called as part of a macro expansion";


MacroProcessorVersion[]=1.01;


Begin["`Private`"];


(* ::Subsection:: *)
(*Initialisations (part of which must not be repeated)*)


version=1.01;
If[! ValueQ[macro\[Breve]substitutions],
   macro\[Breve]substitutions=Hold[];
   executable\[Breve]macros={};
   ];
expanding\[Breve]macros=False;
performing\[Breve]expansion=False;
SetAttributes[md\[Breve]hold,HoldAll];
SetAttributes[macro\[Breve]expand,HoldAll];
passivate[y_]:=(Unevaluated[y]/. x_Symbol:>md\[Breve]hold[x]);

activate[y_]:=(y/.md\[Breve]hold[x_]->x);


(* ::Subsection:: *)
(*Macro\[Breve]Hold and related functions*)


SetAttributes[Macro\[Breve]Hold,HoldAll];
Macro\[Breve]Hold[x_]:=x/;(!performing\[Breve]expansion);


Performing\[Breve]Macro\[Breve]Expansion[]:=performing\[Breve]expansion;


(* ::Subsection:: *)
(*Function called by :=  macro\[Breve]expand*)


macro\[Breve]expand[a_,b_,tag_]:=Module[{tmp,mh1},Block[{performing\[Breve]expansion=True},
   tmp=(Hold[b]/.Macro\[Breve]Hold->mh1//.List@@macro\[Breve]substitutions)//. md\[Breve]hold[x_]->x;
   tmp=passivate[tmp];
   tmp=tmp/.Map[(md\[Breve]hold[#][z___]:>passivate[activate[md\[Breve]hold[#][z]]])&,executable\[Breve]macros];
   tmp=tmp/.{
      (* This list of code tidying up operations can obviously be expanded at will *)
      md\[Breve]hold[Plus][y__]:>simplify\[Breve]plus[md\[Breve]hold[Plus][y]],
         md\[Breve]hold[Times][y__]:>simplify\[Breve]times[md\[Breve]hold[Times][y]],
         md\[Breve]hold[StringJoin][y__String]:>StringJoin[y],
         md\[Breve]hold[StringLength][y_String]:>StringLength[y]
      };
   tmp=activate[tmp]/.Macro\[Breve]Hold[x_]->x /. mh1->Macro\[Breve]Hold;
];
   Switch[tag,

   0,SetDelayed@@Prepend[tmp,Unevaluated[a]],

   1,UpSetDelayed@@Prepend[tmp,Unevaluated[a]],

   _,TagSetDelayed@@Prepend[Prepend[tmp,Unevaluated[a]],tag]
   ];
   ];


(* ::Subsection:: *)
(*Hook of SetDelayed etc.*)


Unprotect[SetDelayed];
SetDelayed[a_,b_]:=Block[{expanding\[Breve]macros=False},macro\[Breve]expand[a,b,0]]/;expanding\[Breve]macros;
Protect[SetDelayed];

Unprotect[UpSetDelayed];
UpSetDelayed[a_,b_]:=Block[{expanding\[Breve]macros=False},macro\[Breve]expand[a,b,1]]/;expanding\[Breve]macros;
Protect[UpSetDelayed];

Unprotect[TagSetDelayed];
TagSetDelayed[f_,a_,b_]:=Block[{expanding\[Breve]macros=False},macro\[Breve]expand[a,b,f]]/;expanding\[Breve]macros;
Protect[TagSetDelayed];


(* ::Subsection:: *)
(*Define\[Breve]Macro function*)


SetAttributes[Define\[Breve]Macro,HoldAll];
Define\[Breve]Macro::called="The macro `` was used other outside a delayed definition";
Options[Define\[Breve]Macro]={"Strict"->True};
Define\[Breve]Macro[x_Symbol]:=(executable\[Breve]macros=Append[executable\[Breve]macros,x]);
Define\[Breve]Macro[mac_,val_,OptionsPattern[Define\[Breve]Macro]]:=Block[{expanding\[Breve]macros=False},
   macro\[Breve]substitutions=Append[macro\[Breve]substitutions,Unevaluated[mac->md\[Breve]hold[val]]];
   If[OptionValue["Strict"],
      With[{tmp=Head[mac]},
      mac:=Message[Define\[Breve]Macro::called,tmp]/;(!performing\[Breve]expansion);
      ],
      mac:=val/;(!performing\[Breve]expansion);      
    ];
   ];

simplify\[Breve]plus[x_]:=Module[{},
   Plus@@Replace[Sort[List@@x],{Longest[y__?NumberQ],z___}:>md\[Breve]hold[Plus][Plus[y],z]]
   ];

simplify\[Breve]times[x_]:=Module[{},
   Times@@Replace[Sort[List@@x],{Longest[y__?NumberQ],z___}:>md\[Breve]hold[Times][Times[y],z]]
   ];
expanding\[Breve]macros=True;


End[];


EndPackage[];
