(* Content-type: application/mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 7.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       145,          7]
NotebookDataLength[     26294,        805]
NotebookOptionsPosition[     20652,        631]
NotebookOutlinePosition[     23083,        702]
CellTagsIndexPosition[     22916,        695]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["MacroProcessor package", "Chapter",
 CellChangeTimes->{{3.4422352966623373`*^9, 3.442235300928472*^9}, {
   3.44731239346875*^9, 3.447312399453125*^9}, {3.447390701375*^9, 
   3.447390712625*^9}, {3.447390810796875*^9, 3.4473908173125*^9}, {
   3.4523472274375*^9, 3.452347231625*^9}, {3.4996253268141565`*^9, 
   3.4996253368297815`*^9}, {3.4996255731735315`*^9, 3.4996255813610315`*^9}, 
   3.499889488421875*^9, 3.50018953578125*^9, 3.500189693171875*^9, {
   3.50018983459375*^9, 3.50018983990625*^9}}],

Cell["\<\
Many C and C++ programmers make extensive use of macros. These enable the \
programmer to write code that is changed even while it is being compiled. \
Frequently this is done to make code more efficient without making it obscure \
to read. For example, one might define a macro such as:\
\>", "Text",
 CellChangeTimes->{{3.500188910140625*^9, 3.500188925171875*^9}, 
   3.500189554671875*^9, 3.5001897096875*^9, 3.500189870125*^9}],

Cell[BoxData[
 RowBox[{"#", "define", " ", "next_char", 
  RowBox[{"(", ")"}], " ", 
  RowBox[{"charpt", "++"}]}]], "Code",
 InitializationCell->False,
 CellChangeTimes->{{3.490449927984375*^9, 3.49044993846875*^9}, {
  3.4904499923125*^9, 3.490450161453125*^9}, {3.4904520784375*^9, 
  3.490452206890625*^9}, {3.50018990071875*^9, 3.500189902015625*^9}}],

Cell["\<\
This would mean that the programmer would write next_char(), but instead of \
this compiling into a function call, it would be replaced at compile time by \
the code \[OpenCurlyDoubleQuote]charpt++\[OpenCurlyDoubleQuote], which \
increments a pointer (usually one machine instruction).\
\>", "Text",
 CellChangeTimes->{{3.5001890845*^9, 3.5001891330625*^9}, 3.50018919646875*^9,
    3.50018993521875*^9}],

Cell[TextData[{
 "This package allows you to do the same thing in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " code. It defines macros that are applied to the right hand side of any \
delayed assignment. They offer similar benefits to C macros, but with much \
less opportunity to write obscure code! Indeed, one of the main reasons to \
use C macros, is to write elegant code that nevertheless executes efficiently."
}], "Text",
 CellChangeTimes->{3.500189951828125*^9}],

Cell["\<\
There are three types of delayed assignment, all of which are can expand \
macros:\
\>", "Text",
 CellChangeTimes->{{3.5001899670625*^9, 3.500189984609375*^9}, {
  3.500308708875*^9, 3.500308709546875*^9}}],

Cell["\[FilledSmallCircle] SetDelayed", "Text",
 CellChangeTimes->{3.50019000128125*^9}],

Cell[BoxData[
 RowBox[{"a", ":=", "b"}]], "Input",
 CellChangeTimes->{{3.5001900341875*^9, 3.500190037109375*^9}}],

Cell["\[FilledSmallCircle] UpSetDelayed", "Text",
 CellChangeTimes->{3.5001900135625*^9}],

Cell[BoxData[
 RowBox[{"a", "^:=", "b"}]], "Input",
 CellChangeTimes->{{3.50019004403125*^9, 3.500190048984375*^9}}],

Cell["\[FilledSmallCircle] TagSetDelayed", "Text",
 CellChangeTimes->{3.50019002934375*^9}],

Cell[BoxData[
 RowBox[{"f", "/:", "a", ":=", "b"}]], "Input",
 CellChangeTimes->{{3.5001900524375*^9, 3.500190057515625*^9}}],

Cell[CellGroupData[{

Cell["\<\
Substitutional macros \[Dash] a trivial example\
\>", "Section",
 CellChangeTimes->{3.50019010253125*^9}],

Cell["\<\
Suppose we have data about people stored in the form of records with the \
format:
Person[name1, name2,age,sex,.....]\
\>", "Text",
 CellChangeTimes->{3.50019012771875*^9}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"person5", "=", 
  RowBox[{"Person", "[", 
   RowBox[{
   "\"\<Fred\>\"", ",", " ", "\"\<Smith\>\"", ",", " ", "23", ",", " ", 
    "\"\<M\>\""}], "]"}]}]], "Input",
 CellChangeTimes->{{3.490455639203125*^9, 3.490455661421875*^9}, 
   3.49045577090625*^9, {3.4904558244375*^9, 3.49045600521875*^9}, 
   3.49045616534375*^9, {3.49984645540625*^9, 3.4998464964375*^9}, {
   3.49984655125*^9, 3.499846553421875*^9}, {3.499869483078125*^9, 
   3.499869485546875*^9}, {3.499869641859375*^9, 3.49986966384375*^9}},
 CellLabel->"In[2]:="],

Cell[BoxData[
 RowBox[{"Person", "[", 
  RowBox[{"\<\"Fred\"\>", ",", "\<\"Smith\"\>", ",", "23", ",", "\<\"M\"\>"}],
   "]"}]], "Output",
 CellChangeTimes->{3.50019014921875*^9},
 CellLabel->"Out[2]="]
}, Open  ]],

Cell["\<\
An elegant way to access one particular field  \[Dash] say age \[Dash] in \
such a record, might be to write:\
\>", "Text",
 CellChangeTimes->{3.500190169921875*^9}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"ageOf", "[", "x_Person", "]"}], ":=", 
   RowBox[{"x", "[", 
    RowBox[{"[", "3", "]"}], "]"}]}], ";"}]], "Input",
 CellLabel->"In[3]:="],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"adultQ", "[", "x_Person", "]"}], ":=", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"ageOf", "[", "x", "]"}], ">=", "18"}], ")"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.499846790640625*^9, 3.49984683346875*^9}, {
  3.4998693013125*^9, 3.499869313734375*^9}},
 CellLabel->"In[3]:="],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ageOf", "[", "person5", "]"}]], "Input",
 CellChangeTimes->{3.499846566671875*^9},
 CellLabel->"In[4]:="],

Cell[BoxData["23"], "Output",
 CellChangeTimes->{3.499846571046875*^9, 3.499869690625*^9, 3.499871340625*^9,
   3.499881383265625*^9},
 CellLabel->"Out[4]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"adultQ", "[", "person5", "]"}]], "Input",
 CellChangeTimes->{{3.4998468450625*^9, 3.499846851359375*^9}, 
   3.4998693194375*^9},
 CellLabel->"In[5]:="],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.499846852359375*^9, 3.499869693*^9, 3.499871343328125*^9,
   3.499881385328125*^9},
 CellLabel->"Out[5]="]
}, Open  ]],

Cell["\<\
The efficient approach is far more opaque, and also loses flexibility \[Dash] \
it would be very hard to change the structure of Person records at a later \
date.\
\>", "Text",
 CellChangeTimes->{{3.500190260734375*^9, 3.50019029284375*^9}}],

Cell["\<\
By defining a macro, it is possible to obtain efficient code that is also \
easy to read.\
\>", "Text",
 CellChangeTimes->{3.5001903299375*^9}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "ageOf", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Clear", "[", "adultQ", "]"}], ";"}]}], "Input",
 CellChangeTimes->{{3.4998692699375*^9, 3.499869293078125*^9}, {
  3.49986932675*^9, 3.499869330453125*^9}},
 CellLabel->"In[4]:="],

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<MacroProcessor`\>\"", "]"}]], "Input",
 CellLabel->"In[6]:="],

Cell[BoxData[
 RowBox[{"Define\[Breve]Macro", "[", 
  RowBox[{
   RowBox[{"ageOf", "[", "x_", "]"}], ",", 
   RowBox[{"x", "[", 
    RowBox[{"[", "3", "]"}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.499869341015625*^9, 3.499869389671875*^9}},
 CellLabel->"In[7]:="],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"adultQ", "[", "x_Person", "]"}], ":=", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"ageOf", "[", "x", "]"}], ">=", "18"}], ")"}]}], ";"}]], "Input",
 CellLabel->"In[8]:="],

Cell["\<\
After these definitions, the ageOf macro is scpliced into any expression on \
the right hand side of a delayed assignment at any depth\
\>", "Text",
 CellChangeTimes->{3.500190374921875*^9}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "adultQ"}]], "Input",
 CellChangeTimes->{{3.500190397234375*^9, 3.500190403765625*^9}},
 CellLabel->"In[9]:="],

Cell["Global`adultQ", "Print", "PrintUsage",
 CellChangeTimes->{3.5001904060625*^9},
 CellTags->"Info3500190405-5263392"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"adultQ", "[", "x_Person", "]"}], ":=", 
         RowBox[{
          RowBox[{"x", "\[LeftDoubleBracket]", "3", "\[RightDoubleBracket]"}],
           "\[GreaterEqual]", "18"}]}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`adultQ],
  Editable->False]], "Print",
 CellChangeTimes->{3.50019040609375*^9},
 CellTags->"Info3500190405-5263392"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"adultQ", "[", "person5", "]"}]], "Input",
 CellChangeTimes->{{3.499869443859375*^9, 3.499869451828125*^9}},
 CellLabel->"In[14]:="],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{
  3.499869452640625*^9, 3.4998695059375*^9, 3.49986973809375*^9, {
   3.499880474796875*^9, 3.4998805011875*^9}, 3.49988142890625*^9},
 CellLabel->"Out[14]="]
}, Open  ]],

Cell["\<\
\[FilledSmallCircle] Note that the macro is not available outside of a \
delayed function definition, and that unlike a normal function, the macro \
must be defined before the delayed definition that uses it.\
\>", "Text",
 CellChangeTimes->{3.50019044196875*^9}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ageOf", "[", "person5", "]"}]], "Input",
 CellChangeTimes->{{3.50019045521875*^9, 3.500190461046875*^9}},
 CellLabel->"In[10]:="],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"Define\[Breve]Macro", "::", "called"}], "MessageName"], ":", 
  " ", "\<\"The macro \[NoBreak]\\!\\(ageOf\\)\[NoBreak] was used other \
outside a delayed definition\"\>"}]], "Message", "MSG",
 CellChangeTimes->{3.500190462078125*^9},
 CellLabel->"During evaluation of In[10]:="]
}, Open  ]],

Cell["\<\
\[FilledSmallCircle] Although the error message is valuable to alert the user \
to the fact that a macro is being used in a way that precludes actual inline \
expansion, sometimes it is convenient to define a macro that acts as an \
ordinary function outside of delayed functions:\
\>", "Text",
 CellChangeTimes->{3.500190483421875*^9}],

Cell[BoxData[
 RowBox[{"Define\[Breve]Macro", "[", 
  RowBox[{
   RowBox[{"ageOf1", "[", "x_", "]"}], ",", 
   RowBox[{"x", "[", 
    RowBox[{"[", "3", "]"}], "]"}], ",", 
   RowBox[{"\"\<Strict\>\"", "->", "False"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.49988155134375*^9, 3.49988156065625*^9}},
 CellLabel->"In[16]:="],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ageOf1", "[", "person5", "]"}]], "Input",
 CellChangeTimes->{{3.49988156775*^9, 3.499881577703125*^9}},
 CellLabel->"In[17]:="],

Cell[BoxData["23"], "Output",
 CellChangeTimes->{3.499881579640625*^9},
 CellLabel->"Out[17]="]
}, Open  ]],

Cell["\<\
\[FilledSmallCircle] The semantics of a macro are subtly different from those \
of a function:\
\>", "Text",
 CellChangeTimes->{3.500190545609375*^9}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"SetAttributes", "[", 
   RowBox[{"setage", ",", "HoldFirst"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"setage", "[", 
    RowBox[{"x_", ",", "y_"}], "]"}], ":=", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"ageOf", "[", "x", "]"}], "=", "y"}], ")"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.499881703640625*^9, 3.499881782609375*^9}},
 CellLabel->"In[19]:="],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"setage", "[", 
  RowBox[{"person5", ",", "42"}], "]"}]], "Input",
 CellChangeTimes->{{3.499881789640625*^9, 3.499881801328125*^9}},
 CellLabel->"In[21]:="],

Cell[BoxData["42"], "Output",
 CellChangeTimes->{3.49988180209375*^9},
 CellLabel->"Out[21]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData["person5"], "Input",
 CellChangeTimes->{{3.49988180478125*^9, 3.499881806703125*^9}},
 CellLabel->"In[22]:="],

Cell[BoxData[
 RowBox[{"Person", "[", 
  RowBox[{"\<\"Fred\"\>", ",", "\<\"Smith\"\>", ",", "42", ",", "\<\"M\"\>"}],
   "]"}]], "Output",
 CellChangeTimes->{3.49988180828125*^9},
 CellLabel->"Out[22]="]
}, Open  ]],

Cell["\<\
This behaviour is easy to understand by looking at the definition of setage:\
\>", "Text",
 CellChangeTimes->{3.50019058475*^9}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "setage"}]], "Input",
 CellChangeTimes->{{3.49988186084375*^9, 3.499881862390625*^9}},
 CellLabel->"In[23]:="],

Cell["Global`setage", "Print", "PrintUsage",
 CellChangeTimes->{3.49988186359375*^9},
 CellTags->"Info3499881863-3785941"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {
     RowBox[{
      RowBox[{"Attributes", "[", "setage", "]"}], "=", 
      RowBox[{"{", "HoldFirst", "}"}]}]},
    {" "},
    {GridBox[{
       {
        RowBox[{
         RowBox[{"setage", "[", 
          RowBox[{"x_", ",", "y_"}], "]"}], ":=", 
         RowBox[{
          RowBox[{"x", "\[LeftDoubleBracket]", "3", "\[RightDoubleBracket]"}],
           "=", "y"}]}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`setage],
  Editable->False]], "Print",
 CellChangeTimes->{3.499881863625*^9},
 CellTags->"Info3499881863-3785941"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Executable macros", "Section",
 CellChangeTimes->{3.500190618734375*^9}],

Cell["\<\
When Define\[Breve]Macro is used with one argument, that argument is \
evaluated at once and the result is inserted into the code. To make this \
useful, the function that expands the macro usually takes the HoldAll \
attribute, and can return an unevaluated result by wrapping it in Macro\
\[Breve]Hold. This wrapper is removed before the code is spliced in. For \
example:\
\>", "Text",
 CellChangeTimes->{3.500190638515625*^9}],

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<MacroProcessor`\>\"", "]"}]], "Input",
 CellLabel->"In[1]:="],

Cell[BoxData[{
 RowBox[{
  RowBox[{"SetAttributes", "[", 
   RowBox[{"elementOf", ",", "HoldAll"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"elementOf", "[", 
    RowBox[{"x_", ",", "str_String"}], "]"}], ":=", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", "ind", "}"}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"ind", "=", 
       RowBox[{
        RowBox[{"Position", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{
           "\"\<name1\>\"", ",", "\"\<name2\>\"", ",", "\"\<age\>\"", ",", 
            "\"\<sex\>\""}], "}"}], ",", 
          RowBox[{"ToLowerCase", "[", "str", "]"}]}], "]"}], "[", 
        RowBox[{"[", 
         RowBox[{"1", ",", "1"}], "]"}], "]"}]}], ";", "\[IndentingNewLine]", 
      RowBox[{"Macro\[Breve]Hold", "@@", 
       RowBox[{"Insert", "[", 
        RowBox[{
         RowBox[{"Hold", "[", 
          RowBox[{"Part", "[", "x", "]"}], "]"}], ",", "ind", ",", 
         RowBox[{"{", 
          RowBox[{"1", ",", "2"}], "}"}]}], "]"}]}]}]}], 
    "\[IndentingNewLine]", "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Define\[Breve]Macro", "[", "elementOf", "]"}], ";"}]}], "Input",
 CellChangeTimes->{{3.49269444465625*^9, 3.49269445940625*^9}, 
   3.492694519875*^9, {3.4926945656875*^9, 3.49269463596875*^9}, {
   3.492694693859375*^9, 3.4926946976875*^9}, {3.492708614046875*^9, 
   3.492708806703125*^9}, {3.492708841625*^9, 3.492708861828125*^9}, {
   3.49270890678125*^9, 3.492708932375*^9}, {3.49270898009375*^9, 
   3.492709049546875*^9}, {3.492709236828125*^9, 3.49270924434375*^9}, {
   3.4927095776875*^9, 3.492709583484375*^9}, {3.49270971696875*^9, 
   3.4927097179375*^9}, {3.492709771859375*^9, 3.49270980384375*^9}, {
   3.49270986275*^9, 3.492709863984375*^9}, {3.49271095240625*^9, 
   3.492710990828125*^9}, {3.492711051265625*^9, 3.49271107621875*^9}, {
   3.49271110990625*^9, 3.49271114465625*^9}, {3.49271117609375*^9, 
   3.49271117759375*^9}, {3.492711210984375*^9, 3.4927112123125*^9}, {
   3.492711559953125*^9, 3.492711602640625*^9}, 3.4927116376875*^9, {
   3.499888143578125*^9, 3.49988815590625*^9}, {3.49988818740625*^9, 
   3.499888197140625*^9}},
 CellLabel->"In[10]:="],

Cell["Use the macro inside a function", "Text",
 CellChangeTimes->{3.500190673203125*^9}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"shaves", "[", "x_", "]"}], ":=", 
   RowBox[{
    RowBox[{
     RowBox[{"elementOf", "[", 
      RowBox[{"x", ",", "\"\<Sex\>\""}], "]"}], "\[Equal]", "\"\<M\>\""}], 
    " ", "&&", " ", 
    RowBox[{
     RowBox[{"elementOf", "[", 
      RowBox[{"x", ",", "\"\<Age\>\""}], "]"}], " ", ">", "16"}]}]}], 
  ";"}]], "Input",
 CellChangeTimes->{{3.492694648234375*^9, 3.4926946691875*^9}, {
   3.492709067578125*^9, 3.49270909709375*^9}, 3.492709421703125*^9, {
   3.4927145971875*^9, 3.49271461809375*^9}, {3.492714648140625*^9, 
   3.4927146795625*^9}},
 CellLabel->"In[5]:="],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "shaves"}]], "Input",
 CellChangeTimes->{{3.4998824029375*^9, 3.499882405109375*^9}},
 CellLabel->"In[6]:="],

Cell["Global`shaves", "Print", "PrintUsage",
 CellChangeTimes->{3.499882831390625*^9},
 CellTags->"Info3499882831-4181374"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"shaves", "[", "x_", "]"}], ":=", 
         RowBox[{
          RowBox[{
           RowBox[{
           "x", "\[LeftDoubleBracket]", "4", "\[RightDoubleBracket]"}], 
           "\[Equal]", "\<\"M\"\>"}], "&&", 
          RowBox[{
           RowBox[{
           "x", "\[LeftDoubleBracket]", "3", "\[RightDoubleBracket]"}], ">", 
           "16"}]}]}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`shaves],
  Editable->False]], "Print",
 CellChangeTimes->{3.49988283140625*^9},
 CellTags->"Info3499882831-4181374"]
}, Open  ]],

Cell["\<\
\[FilledSmallCircle] Notice that elementOf would be quite inefficient as an \
ordinary function, because it involves various string manipulations, but used \
as a macro, it produces optimal code.\
\>", "Text",
 CellChangeTimes->{3.500190704453125*^9}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"person5", "=", 
  RowBox[{"Person", "[", 
   RowBox[{
   "\"\<Fred\>\"", ",", " ", "\"\<Smith\>\"", ",", " ", "23", ",", " ", 
    "\"\<M\>\""}], "]"}]}]], "Input",
 CellChangeTimes->{{3.490455639203125*^9, 3.490455661421875*^9}, 
   3.49045577090625*^9, {3.4904558244375*^9, 3.49045600521875*^9}, 
   3.49045616534375*^9, {3.49984645540625*^9, 3.4998464964375*^9}, {
   3.49984655125*^9, 3.499846553421875*^9}, {3.499869483078125*^9, 
   3.499869485546875*^9}, {3.499869641859375*^9, 3.49986966384375*^9}},
 CellLabel->"In[7]:="],

Cell[BoxData[
 RowBox[{"Person", "[", 
  RowBox[{"\<\"Fred\"\>", ",", "\<\"Smith\"\>", ",", "23", ",", "\<\"M\"\>"}],
   "]"}]], "Output",
 CellChangeTimes->{3.499882556109375*^9, 3.499882836609375*^9},
 CellLabel->"Out[7]="]
}, Open  ]],

Cell["\<\
\[FilledSmallCircle] Executable macros can be coded so as to run outside \
delayed constructs. They can also call \
Performing\[Breve]Macro\[Breve]Expansion[] to determine on the fly if they \
are performing a macro expansion or acting as a conventional function.\
\>", "Text",
 CellChangeTimes->{3.500190738765625*^9}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"elementOf", "[", 
  RowBox[{"person5", ",", "\"\<Sex\>\""}], "]"}]], "Input",
 CellChangeTimes->{3.499882585359375*^9},
 CellLabel->"In[13]:="],

Cell[BoxData["\<\"M\"\>"], "Output",
 CellChangeTimes->{3.499882586078125*^9, 3.499882839265625*^9, 
  3.4998882089375*^9},
 CellLabel->"Out[13]="]
}, Open  ]]
}, Open  ]]
}, Open  ]]
},
Saveable->False,
Selectable->True,
WindowSize->{717, 741},
WindowMargins->{{Automatic, -1}, {48, Automatic}},
WindowFrame->"Normal",
Visible->True,
DockedCells->FEPrivate`FrontEndResource["FEExpressions", "HelpViewerToolbar"],
PrintingCopies->1,
PrintingPageRange->{Automatic, Automatic},
PageHeaders->{{Inherited, Inherited, 
   Cell[
    TextData[{"The Super Widget Package"}], "PageHeader"]}, {
   Cell[
    TextData[{"The Super Widget Package"}], "PageHeader"], Inherited, 
   Inherited}},
PageFooters->{{Inherited, 
   Cell[
    TextData[{"\[Copyright]2006 David Bailey"}], "PageFooter"], Inherited}, {
  Inherited, 
   Cell[
    TextData[{"\[Copyright]2006 David Bailey"}], "PageFooter"], Inherited}},
PrivateNotebookOptions->{"ShortBoxForm"->False},
ShowCellBracket->True,
ShowSelection->True,
TaggingRules->{
 "ModificationHighlight" -> False, 
  "Metadata" -> {
   "context" -> "DebugTrace`", "keywords" -> {}, "index" -> True, "label" -> 
    "DebugTrace Package Paclet Symbol", "language" -> "en", "paclet" -> 
    "DebugTrace Package", "status" -> "", "summary" -> 
    "DTLocalValue[n] \[Dash] Returns the value of the most recently created \
local variable of the form n$<number>", "synonyms" -> {}, "title" -> 
    "DTLocalValue", "type" -> "Symbol", "uri" -> 
    "DebugTrace/ref/DTLocalValue"}, "SearchTextTranslated" -> "", 
  "SearchText" -> ""},
ShowCellLabel->False,
PrivateNotebookOptions -> {"NotebookAuthor" -> ""},
Magnification->1.,
FrontEndVersion->"7.0 for Microsoft Windows (64-bit) (February 18, 2009)",
StyleDefinitions->"HelpBrowser.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "Info3500190405-5263392"->{
  Cell[7844, 243, 121, 2, 70, "Print",
   CellTags->"Info3500190405-5263392"],
  Cell[7968, 247, 925, 25, 70, "Print",
   CellTags->"Info3500190405-5263392"]},
 "Info3499881863-3785941"->{
  Cell[12645, 402, 122, 2, 70, "Print",
   CellTags->"Info3499881863-3785941"],
  Cell[12770, 406, 1063, 31, 70, "Print",
   CellTags->"Info3499881863-3785941"]},
 "Info3499882831-4181374"->{
  Cell[17620, 539, 123, 2, 70, "Print",
   CellTags->"Info3499882831-4181374"],
  Cell[17746, 543, 1097, 31, 70, "Print",
   CellTags->"Info3499882831-4181374"]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"Info3500190405-5263392", 22349, 677},
 {"Info3499881863-3785941", 22538, 682},
 {"Info3499882831-4181374", 22730, 687}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[567, 22, 512, 7, 85, "Chapter"],
Cell[1082, 31, 442, 7, 55, "Text"],
Cell[1527, 40, 355, 7, 47, "Code",
 InitializationCell->False],
Cell[1885, 49, 414, 7, 39, "Text"],
Cell[2302, 58, 481, 9, 55, "Text"],
Cell[2786, 69, 216, 5, 23, "Text"],
Cell[3005, 76, 88, 1, 23, "Text"],
Cell[3096, 79, 114, 2, 27, "Input"],
Cell[3213, 83, 89, 1, 23, "Text"],
Cell[3305, 86, 116, 2, 27, "Input"],
Cell[3424, 90, 91, 1, 23, "Text"],
Cell[3518, 93, 125, 2, 27, "Input"],
Cell[CellGroupData[{
Cell[3668, 99, 115, 3, 63, "Section"],
Cell[3786, 104, 182, 5, 47, "Text"],
Cell[CellGroupData[{
Cell[3993, 113, 554, 11, 27, "Input"],
Cell[4550, 126, 202, 5, 25, "Output"]
}, Open  ]],
Cell[4767, 134, 175, 4, 23, "Text"],
Cell[4945, 140, 185, 6, 27, "Input"],
Cell[5133, 148, 330, 9, 27, "Input"],
Cell[CellGroupData[{
Cell[5488, 161, 129, 3, 27, "Input"],
Cell[5620, 166, 157, 3, 25, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[5814, 174, 176, 4, 27, "Input"],
Cell[5993, 180, 159, 3, 70, "Output"]
}, Open  ]],
Cell[6167, 186, 251, 5, 70, "Text"],
Cell[6421, 193, 153, 4, 70, "Text"],
Cell[6577, 199, 296, 7, 70, "Input"],
Cell[6876, 208, 103, 2, 70, "Input"],
Cell[6982, 212, 272, 7, 70, "Input"],
Cell[7257, 221, 218, 7, 70, "Input"],
Cell[7478, 230, 200, 4, 70, "Text"],
Cell[CellGroupData[{
Cell[7703, 238, 138, 3, 70, "Input"],
Cell[7844, 243, 121, 2, 70, "Print",
 CellTags->"Info3500190405-5263392"],
Cell[7968, 247, 925, 25, 70, "Print",
 CellTags->"Info3500190405-5263392"]
}, Open  ]],
Cell[CellGroupData[{
Cell[8930, 277, 155, 3, 70, "Input"],
Cell[9088, 282, 210, 4, 70, "Output"]
}, Open  ]],
Cell[9313, 289, 273, 5, 70, "Text"],
Cell[CellGroupData[{
Cell[9611, 298, 153, 3, 70, "Input"],
Cell[9767, 303, 326, 7, 70, "Message"]
}, Open  ]],
Cell[10108, 313, 346, 6, 70, "Text"],
Cell[10457, 321, 324, 8, 70, "Input"],
Cell[CellGroupData[{
Cell[10806, 333, 151, 3, 70, "Input"],
Cell[10960, 338, 95, 2, 70, "Output"]
}, Open  ]],
Cell[11070, 343, 160, 4, 70, "Text"],
Cell[11233, 349, 427, 12, 70, "Input"],
Cell[CellGroupData[{
Cell[11685, 365, 179, 4, 70, "Input"],
Cell[11867, 371, 94, 2, 70, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[11998, 378, 122, 2, 70, "Input"],
Cell[12123, 382, 203, 5, 70, "Output"]
}, Open  ]],
Cell[12341, 390, 138, 3, 70, "Text"],
Cell[CellGroupData[{
Cell[12504, 397, 138, 3, 70, "Input"],
Cell[12645, 402, 122, 2, 70, "Print",
 CellTags->"Info3499881863-3785941"],
Cell[12770, 406, 1063, 31, 70, "Print",
 CellTags->"Info3499881863-3785941"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[13882, 443, 78, 1, 70, "Section"],
Cell[13963, 446, 440, 8, 70, "Text"],
Cell[14406, 456, 103, 2, 70, "Input"],
Cell[14512, 460, 2227, 48, 70, "Input"],
Cell[16742, 510, 89, 1, 70, "Text"],
Cell[16834, 513, 622, 17, 70, "Input"],
Cell[CellGroupData[{
Cell[17481, 534, 136, 3, 70, "Input"],
Cell[17620, 539, 123, 2, 70, "Print",
 CellTags->"Info3499882831-4181374"],
Cell[17746, 543, 1097, 31, 70, "Print",
 CellTags->"Info3499882831-4181374"]
}, Open  ]],
Cell[18858, 577, 261, 5, 70, "Text"],
Cell[CellGroupData[{
Cell[19144, 586, 554, 11, 70, "Input"],
Cell[19701, 599, 225, 5, 70, "Output"]
}, Open  ]],
Cell[19941, 607, 329, 6, 70, "Text"],
Cell[CellGroupData[{
Cell[20295, 617, 167, 4, 70, "Input"],
Cell[20465, 623, 147, 3, 70, "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)

